#!/usr/bin/perl
# Copyright (C) 2017 Yuri V. Kravatsky, jiri@eimb.ru - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the Creative Commons CC BY-NC-SA 3.0 license.
# If you use this script in your scienfic projects please cite:
# Kravatsky YV, Chechetkin VR, Fedoseeva DM, Gorbacheva MA, Kravatskaya GI, Kretova OV, Tchurikov NA.
# A bioinformatic pipeline for monitoring of the mutational stability of viral drug targets with deep-sequencing technology.
# DOI: XX.XXXX/XXXX, PMID: XXXXXX

use Bio::SeqIO;
use Bio::SearchIO;
use File::Copy;

### This is FASTA parameters. Tune it according to your needs. In the most cases you could want to change
### only E-value that corresponds to alignment quality. Look to FASTA manual for further details
    my $options = "-3 -E 0.01 -q -n";
    my $FASTANAME="ssearch35";

    my $ARGC=scalar @ARGV;
    if ($ARGC!=6) { die "Usage: find_invariable aligned.fastafile targetfile outfile targetlength numruns RAMdrivepath\n"; }

    my $aligned=$ARGV[0];
    my $targetfile=$ARGV[1];
    my $outputfile=$ARGV[2];
    my $TARGETLENGTH=$ARGV[3];
    my $NUMRUNS=$ARGV[4];
    my $RAMDRIVE= $ARGV[5] . "/";

    open (TARG,"$targetfile") || die "Can't read from \"$targetfile\": $!";
    open (OUTP,">$outputfile") || die "Can't write to \"$outputfile\": $!";

    my $target;

    while(<TARG>) {
       $target = $target . $_;
                  }
    close (TARG);
    print "$target";

    my $seqio = Bio::SeqIO->new( -file => $aligned,
                                 -format => "fasta",
                                 -alphabet  => "dna" );
    my %targets;
    my $totaltargets=0;

# Copy to RAMdrive ssearch and target to speedup FASTA runs
    my $fasta   = $RAMDRIVE . $FASTANAME;
    my $library = $RAMDRIVE . $aligned;
    my $ftarg = $RAMDRIVE . $targetfile;
    if (-e $ftarg) { unlink $ftarg; }
    if (copy ($targetfile, $ftarg)==0) { die("Cannot copy target file to RAMdrive: $!\n"); }
    my $totalrecords=0;

nextrecord: while (my $fastaseq = $seqio->next_seq) {
#write temporary file with FASTA database

                  my $id=$fastaseq->display_id;
                  my @head=split(/\./,$id);
                  my $numreads=$head[1];
                   my $seq_out = Bio::SeqIO->new(-file   => ">$library",
                                                 -format => "fasta",
                                                 -alpabet => "dna" );
                   $seq_out->write_seq($fastaseq);
# create command line to run FASTA search
                   my $command = "$fasta $options $ftarg $library";
# run FASTA search, capture output
                   open $fh,"$command |" || die("Cannot run fasta cmd of $command: $!\n");
                   my $in  = Bio::SearchIO->new(-format => 'fasta', -fh => $fh);
                   while( my $result = $in->next_result ) {
                       while( my $hit = $result->next_hit ) {
## $hit is a Bio::Search::Hit::HitI compliant object
                           while( my $hsp = $hit->next_hsp ) {
## $hsp is a Bio::Search::HSP::HSPI compliant object
                               $query_str = $hsp->seq_str('query');
                               $hit_str = $hsp->seq_str('hit');

                               my ($query_beg, $query_end) = $hsp->range('query');
                               my ($hit_beg, $hit_end) = $hsp->range('hit');
                               $query_str =~ s/^( *)//g;
                               my $offset=length($1);
                               my $leng=length($query_str);
                               my $final=substr $hit_str, $offset, $leng;
                               $final =~ s/^( *)//g;
                               if (length($final)<$TARGETLENGTH) { goto nextrecord; }
                               $targets { $final } += $numreads;
                               $totaltargets += $numreads;
                               $totalrecords++;

                               if ($NUMRUNS) {        # if $NUMRUNS==0 run all records up to the end of file
                                   if ($totalrecords>=$NUMRUNS) {
                                          print "Smith-Waterman alignments $NUMRUNS evaluation runs fishined, restarting pipeline with refined target\n";
                                          goto fin;             }
                                             }
                               goto nextrecord;              }
                                                            }
                                                           }
                                            }
fin:    print "$totaltargets targets processed\n";

        my $totalhits=0;
        foreach my $targvalue (sort { $targets{$b} <=> $targets{$a} } keys %targets) {
            $totalhits += $targets{$targvalue};
            my $perc=100.0 * $targets{$targvalue} / $totaltargets;
            my $c=sprintf "%s\t%d\t%.3f\n", $targvalue, $targets{$targvalue}, $perc;
            print OUTP $c;                                                           }
        print OUTP "Total $totalhits hits found\n";
        close (OUTP);
