#!/usr/bin/perl
# Copyright (C) 2017 Yuri V. Kravatsky, jiri@eimb.ru - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the Creative Commons CC BY-NC-SA license.
# If you use this script in your scienfic projects please cite:
# Kravatsky YV, Chechetkin VR, Fedoseeva DM, Gorbacheva MA, Kravatskaya GI, Kretova OV, Tchurikov NA.
# Bioinformatic pipeline for deep sequencing monitoring of mutational stability of viral drug targets. 
# DOI: XX.XXXX/XXXX, PMID: XXXXXX

    use strict;
    my $THRESHOLD=2; # All alignments that happened less than THRESHOLD times will be omitted from the future analysis
    my $unpackflag=0;
    my $statflag=0;
    my $logoflag=0;

    my $ARGC=scalar @ARGV;

    if ($ARGC<1 || $ARGC>2) { die "Usage: result_to_fasta resultfile (unpack) (stat) (logo)\n"; }

    my $infile=$ARGV[0];
    my $basename=substr($infile,0,rindex($infile,"."));

    open (INP,"$infile") || die "Can't read from \"$infile\": $!";
    my $option=uc($ARGV[1]);
    if ($ARGC==2) { if ($option eq "UNPACK") { $unpackflag=1; }
                      elsif ($option eq "STAT") { $statflag=1; }
                      elsif ($option eq "LOGO") { $logoflag=1; }
                         else { die "Unknown option $ARGV[1]"; }
                  }

    my $outfile;
   if ($logoflag) { $outfile=$basename."_logo.fa"; }
       else       { $outfile=$basename.".fa"; }

    open (OUTP,">$outfile") || die "Can't write to \"$outfile\": $!";

    my $count=1;
    my $prevheader;
    my $lastcycle=0;
#  Read the first string with the most invariable target
    my $mostinv=<INP>;
    my @mostdat=split('\t',$mostinv);
    my $logolength=length($mostdat[0]);
    seek INP, 0, 0;
    while(<INP>) {
        chomp;
        chomp;
        my @arr=split('\t');
        if (scalar @arr!=3) { 
#           @arr=split('\t',$mostinv);
           $lastcycle=1;
           last;
                            }
        my $cycles;
        if ($unpackflag) { $cycles=$arr[1]; }
           elsif ($logoflag) { $cycles=$arr[1]; }
               else          { $cycles=1; }

        if ($statflag || $logoflag) { if ($arr[1]<$THRESHOLD) { next; } }

        $_ = reverse $arr[1];
        s/(\d{3})(?=\d)(?!\d*\.)/$1 /g;
        my $number = reverse $_;

       if ($logoflag) { my $currlength=length($arr[0]);
                        if ($currlength<$logolength) { next; }
                        if ($currlength>$logolength) { $arr[0] = substr($arr[0],0,$logolength); }
                      }

        for (my $i=0;$i<$cycles;$i++) { my $c;
  		                        my $finalheader;
      					if (!$unpackflag) { 
							    if ($statflag) { $finalheader=">$count." ."$arr[1]"; }
                                                                   else    { $c=sprintf ">$number (%.2f)", $arr[2]; 
      					                                     if ($c eq $prevheader) { $finalheader = $c ."\t$count"; }
      					                                                       else { $finalheader=$c; }
                                                                             $prevheader=$c;
                                                                           }
                                                          }
                                                 else     { $finalheader=sprintf ">$count $arr[1] (%.2f)", $arr[2]; }
                                        if ($logoflag) { $finalheader=sprintf ">$count $arr[1] (%.2f)", $arr[2]; }
  				        print OUTP "$finalheader\n$arr[0]\n";
                                        $count++; }
        if ($lastcycle) { last; }
                  }
    close (INP);
    close (OUTP);
