#!/usr/bin/perl
# Copyright (C) 2017 Yuri V. Kravatsky, jiri@eimb.ru - All Rights Reserved
# You may use, distribute and modify this code under the
# terms of the Creative Commons CC BY-NC-SA 3.0 license.
# If you use this script in your scienfic projects please cite:
# Kravatsky YV, Chechetkin VR, Fedoseeva DM, Gorbacheva MA, Kravatskaya GI, Kretova OV, Tchurikov NA.
# A bioinformatic pipeline for monitoring of the mutational stability of viral drug targets with deep-sequencing technology.
# DOI: XX.XXXX/XXXX, PMID: XXXXXX

     use Statistics::Descriptive;
     use BioUtil::Seq;

    my $ARGC=scalar @ARGV;

    if ($ARGC==0||$ARGC>3) { die "Usage: mut_fastastat.pl fasta_aligned.fa\n"; }

    my $infile=$ARGV[0];
    my $basename=substr($infile,0,rindex($infile,"."));
    my $outfile=$basename.".stat";
    open (INP,"$infile") || die "Can't read from \"$infile\": $!";
    close (INP);
    open (OUTP,">$outfile") || die "Can't write to \"$outfile\": $!";

    my $next_seq = FastaReader($infile);

    my $numseq;
    my @seqarr;
    my @seqnum;
    my $i=0;
    my $alignlength;
    my $totalnumseq=0;
    while ( my $fa = &$next_seq() ) {
        my ( $header, $seq ) = @$fa;
        if ($i==0) { $alignlength=length($seq); }
        if (length($seq) != $alignlength) { print STDERR "Warning: Alignment is not box in string $header, string omitted, check output thouroughly!\n"; 
                                            print STDERR "String $seq has length that is not equal to $alignlength\n";
                                            next; }
        my @head=split(/\./,$header);
        my $numstrings;
        if (uc($head[0]) eq "CONSENSUS") { $numstrings=0; }
           else { $numstrings=$head[1];
                  if (!($numstrings =~ /^\d+\z/)) { print "We've got $numstrings as number of strings, ignore\n"; next; }
                }
        $seqarr[$i] = uc($seq);
        $seqnum[$i] = $numstrings;
        $totalnumseq += $numstrings;
        $i++;                       }

    $numseq=$i;

    my @transitions;
    my @transversions;
    my $realnumseq=$numseq-1;
#print "Number of sequences is $realnumseq\n";

    for (my $i=0;$i<$realnumseq;$i++) {
          for (my $j=0;$j<3;$j++) {
               $transitions[$i][$j]=0;
	       $transversions[$i][$j]=0;
       	                          }   }

#### here we should detect effective length (cut ------ at the end)
#       my $reallength=$alignlength;
#       for (my $i=$alignlength;$i>0;$i--) {
#          my $emptychar=0;
#          for (my $j=0;$j<$numseq;$j++) { 
#               if ($seqarr[$j]->subseq($i,$i) eq "-") { $emptychar++; }
#                                        }
#          if ($emptychar<$numseq) { last; } 
#          $reallength=$i;  
#                                          }

#### Cut all "-" at the end - align to block
       my $reallength=$alignlength;
#cuttail: for (my $i=$alignlength;$i>0;$i--) {
#          for (my $j=0;$j<$numseq;$j++) { 
#               if ($seqarr[$j]->subseq($i,$i) eq "-") { next cuttail; }
#                                        }
#          $reallength=$i;  
#          last;                             }

#print OUTP "Detected length=$alignlength, Real length=$reallength\n";

#convert all string to upper character
#       for (my $j=0;$j<$numseq;$j++) {
#          my $seq=$seqarr[$j]->seq;
#          $seqarr[$j]->seq(uc($seq)); 
#                                     }

       my (@delarr,  @insarr);
       my (@AT, @AG, @AC, @TA, @TG, @TC, @GA, @GT, @GC, @CA, @CT, @CG);

       for (my $i=0;$i<$reallength;$i++) {
            $AT[$i]=$AG[$i]=$AC[$i]=$TA[$i]=$TG[$i]=$TC[$i]=$GA[$i]=$GT[$i]=$GC[$i]=$CA[$i]=$CT[$i]=$CG[$i]=0;
                                         }

       for (my $i=0;$i<$reallength;$i++) {
          for (my $j=0;$j<$realnumseq;$j++) {
                   $delarr[$j][$i]=0;
                   $insarr[$j][$i]=0;           }
#              my $numsym=$i+1;
#              $wildsym=$seqarr[$realnumseq]->subseq($numsym,$numsym);
              my $numsym=$i;
              $wildsym=substr($seqarr[$realnumseq],$numsym,1);
              for (my $j=0;$j<$realnumseq;$j++) {
#                    my $currsym=$seqarr[$j]->subseq($numsym, $numsym);
                     my $currsym=substr($seqarr[$j],$numsym,1);
                     if ($currsym eq 'N') { last; }
                     if ($currsym eq $wildsym) { next; }
                    $pos=$i%3;
# my $numseq=$j+1;
# $numpos=$pos+1;
# print OUTP "$numpos:$numsym:$numseq $wildsym => $currsym\n";
# Insertions: Reference "-" => any
                    if (($wildsym eq "-") && ($currsym ne "-")) { $insarr[$j][$i]+=$seqnum[$j];
                                                                  next; }
# Deletions:  Reference any => "-"
                    if (($wildsym ne "-") && ($currsym eq "-")) { $delarr[$j][$i]+=$seqnum[$j]; 
                                                                  next; }
                    if ($wildsym eq "C") {
                         if ($currsym eq "T") { $transitions[$j][$pos]+=$seqnum[$j];
                                                $CT[$i]+=$seqnum[$j];
                                                next; }
                         if ($currsym eq "A") { $transversions[$j][$pos]+=$seqnum[$j];
                                                $CA[$i]+=$seqnum[$j];
                                                next; }
                         if ($currsym eq "G") { $transversions[$j][$pos]+=$seqnum[$j];
                                                $CG[$i]+=$seqnum[$j];
                                                next; }
                                         }
                    if ($wildsym eq "A") {
                         if ($currsym eq "G") { $transitions[$j][$pos]+=$seqnum[$j];
                                                $AG[$i]+=$seqnum[$j];
                                                next; }
                         if ($currsym eq "C") { $transversions[$j][$pos]+=$seqnum[$j];
                                                $AC[$i]+=$seqnum[$j];
                                                next; }
                         if ($currsym eq "T") { $transversions[$j][$pos]+=$seqnum[$j];
                                                $AT[$i]+=$seqnum[$j];
                                                next; }
                                         }
                    if ($wildsym eq "G") {
                         if ($currsym eq "A") { $transitions[$j][$pos]+=$seqnum[$j]; 
                                                $GA[$i]+=$seqnum[$j];
                                                next; }
                         if ($currsym eq "T") { $transversions[$j][$pos]+=$seqnum[$j]; 
                                                $GT[$i]+=$seqnum[$j];
                                                next; }
                         if ($currsym eq "C") { $transversions[$j][$pos]+=$seqnum[$j]; 
                                                $GC[$i]+=$seqnum[$j];
                                                next; }
                                         }
                    if ($wildsym eq "T") {
                         if ($currsym eq "C") { $transitions[$j][$pos]+=$seqnum[$j]; 
                                                $TC[$i]+=$seqnum[$j];
                                                next; }
                         if ($currsym eq "A") { $transversions[$j][$pos]+=$seqnum[$j]; 
                                                $TA[$i]+=$seqnum[$j];
                                                next; }
                         if ($currsym eq "G") { $transversions[$j][$pos]+=$seqnum[$j]; 
                                                $TG[$i]+=$seqnum[$j];
                                                next; }
                                         }
                    print STDERR "Impossible situation, unknown mutation!\n";
                                                }
                                          }

       my @insertions;
       my @deletions;

       for (my $j=0;$j<$realnumseq;$j++) {
          my @currarr;
          for (my $i=0;$i<$reallength;$i++) {
             $currarr[$i]=$insarr[$j][$i];  }
          push (@insertions, indels(\@currarr,$reallength));
          for (my $i=0;$i<$reallength;$i++) {
             $currarr[$i]=$delarr[$j][$i];  }
          push (@deletions, indels(\@currarr,$reallength));
                                         }

       my @transitsum;
       my @transversum;
       my @transitfreq;
       my @transverfreq;
       my @totalsum;
       my @totalfreq;

       my @transitseq;
       my @transverseq;
       my @sumseq;
       for (my $i=0;$i<$realnumseq;$i++) {
           $transitseq[$i]=0;
           $transverseq[$i]=0;
           for (my $j=0;$j<3;$j++) {
             $transitseq[$i]+=$transitions[$i][$j];
             $transverseq[$i]+=$transversions[$i][$j];
                                   }     
           $sumseq[$i]=$transitseq[$i]+$transverseq[$i]; }

       for (my $j=0;$j<3;$j++) {
          $transitsum[$j]=0;
          $transversum[$j]=0;
          for (my $i=0;$i<$realnumseq;$i++) {
              $transitsum[$j]+=$transitions[$i][$j];
              $transversum[$j]+=$transversions[$i][$j];
                                            }
          $transitfreq[$j]=$transitsum[$j]/($reallength*$totalnumseq);
          $transverfreq[$j]=$transversum[$j]/($reallength*$totalnumseq);
          $totalsum[$j]=$transitsum[$j]+$transversum[$j];
          $totalfreq[$j]=$totalsum[$j]/($reallength*$totalnumseq);
                               } 
       my $threshold=3.84/$totalnumseq;
       print OUTP "Sequences/Total\t$totalnumseq\t$reallength\n";
#       print OUTP "Sum by pos\nTotal:\t$totalsum[0]\t$totalsum[1]\t$totalsum[2]\n";
#       print OUTP "Transitions:\t$transitsum[0]\t$transitsum[1]\t$transitsum[2]\n";
#       print OUTP "Transversions:\t$transversum[0]\t$transversum[1]\t$transversum[2]\n";

#       my $c=sprintf("Freq by pos\nTotal:\t%.6f\t%.6f\t%.6f\n",$totalfreq[0],$totalfreq[1],$totalfreq[2]);
#       print OUTP $c;
#       $c=sprintf("Transitions:\t%.6f\t%.6f\t%.6f\nTransversions:\t%.6f\t%.6f\t%.6f\n",$transitfreq[0],$transitfreq[1],$transitfreq[2],$transverfreq[0],$transverfreq[1],$transverfreq[2]);
#       print OUTP $c;

       my $globalsum=$totalsum[0]+$totalsum[1]+$totalsum[2];
       my $globalfreq=$globalsum/($totalnumseq*$reallength);
       $c=sprintf("Mutations Total/Frequency\t$globalsum\t%.5f\n",$globalfreq);
       print OUTP $c;


       my $totalins=scalar @insertions;
       my $totaldel=scalar @deletions;
       my @y=();

       my $statinsert = Statistics::Descriptive::Full->new();
       $statinsert->add_data(@y);

       $statinsert->add_data(@insertions);
       my $median=$statinsert->median();
       my $stddev=$statinsert->standard_deviation();
       my $mean=$statinsert->mean();
       my $min=$statinsert->min();
       my $max=$statinsert->max();

#print OUTP "Insertions\t@insertions\n";

#       my $freqins=$totalins/($reallength*$realnumseq);
#print "Total number of sequences is $totalnumseq, alignment length=$reallength\n";
       my $freqins=$totalins/($reallength*$totalnumseq);
       $c=sprintf("Insertions\t$mean\t$median\t$stddev\t$max\t$min\t$totalins\t%.8f\n",$freqins);
       print OUTP $c;
       undef $statinsert;

       my $statdelete = Statistics::Descriptive::Full->new();
       $statdelete->add_data(@y);

       $statdelete->add_data(@deletions);
       $median=$statdelete->median();
       $stddev=$statdelete->standard_deviation();
       $mean=$statdelete->mean();
       $min=$statdelete->min();
       $max=$statdelete->max();

#print OUTP "Deletions\t@deletions\n";
       my $freqdel=$totaldel/($reallength*$realnumseq);
       $c=sprintf("Deletions\t$mean\t$median\t$stddev\t$max\t$min\t$totaldel\t%.8f\n",$freqdel);
       print OUTP $c;
       undef $statdelete;
##       print OUTP "Transitions/seq";
##       for (my $i=0;$i<$realnumseq;$i++) { print OUTP "\t$transitseq[$i]"; }
##       print OUTP "\nTransversions/seq";
##       for (my $i=0;$i<$realnumseq;$i++) { print OUTP "\t$transverseq[$i]"; }
##       print OUTP "\nMutations/seq";
##       for (my $i=0;$i<$realnumseq;$i++) { print OUTP "\t$sumseq[$i]"; }
##       print OUTP "\n";

       print OUTP "\nMutations per nucleotide position\n";
       print OUTP "\t\tAT\tAG\tAC\tTA\tTG\tTC\tGA\tGT\tGC\tCA\tCT\tCG\tsum\tthreshold\n";

       my @summut;
       for (my $i=0;$i<$reallength;$i++) {
             my $sum=($AT[$i]+$AG[$i]+$AC[$i]+$TA[$i]+$TG[$i]+$TC[$i]+$GA[$i]+$GT[$i]+$GC[$i]+$CA[$i]+$CT[$i]+$CG[$i])/$totalnumseq;
             $summut[$i]=$sum;
             my $pos=$i;
             my $currsym=substr($seqarr[$realnumseq],$pos,1);

             $AT[$i]/=$totalnumseq; 
             $AG[$i]/=$totalnumseq; 
             $AC[$i]/=$totalnumseq; 

             $TA[$i]/=$totalnumseq; 
             $TG[$i]/=$totalnumseq; 
             $TC[$i]/=$totalnumseq; 

             $GA[$i]/=$totalnumseq; 
             $GT[$i]/=$totalnumseq; 
             $GC[$i]/=$totalnumseq; 

             $CA[$i]/=$totalnumseq; 
             $CT[$i]/=$totalnumseq; 
             $CG[$i]/=$totalnumseq; 

             if ($AT[$i]==0) { $AT[$i]=''; }
             if ($AG[$i]==0) { $AG[$i]=''; }
             if ($AC[$i]==0) { $AC[$i]=''; }

             if ($TA[$i]==0) { $TA[$i]=''; }
             if ($TG[$i]==0) { $TG[$i]=''; }
             if ($TC[$i]==0) { $TC[$i]=''; }

             if ($GA[$i]==0) { $GA[$i]=''; }
             if ($GT[$i]==0) { $GT[$i]=''; }
             if ($GC[$i]==0) { $GC[$i]=''; }

             if ($CA[$i]==0) { $CA[$i]=''; }
             if ($CT[$i]==0) { $CT[$i]=''; }
             if ($CG[$i]==0) { $CG[$i]=''; }

             $pos=$i+1;
             print OUTP "$pos\t$currsym\t$AT[$i]\t$AG[$i]\t$AC[$i]\t$TA[$i]\t$TG[$i]\t$TC[$i]\t$GA[$i]\t$GT[$i]\t$GC[$i]\t$CA[$i]\t$CT[$i]\t$CG[$i]\t$sum\t$threshold\n";
                                         }

       my @y=();
       my $stat = Statistics::Descriptive::Full->new();
       $stat->add_data(@y);
       $stat->add_data(@summut);
       my $median=$stat->median();
       my $mean=$stat->mean();
       my $stddev=$stat->standard_deviation();
       my $minimum=$stat->min();
       my $maximum=$stat->max();
       my $lower=$stat->quantile(1);
       my $upper=$stat->quantile(3);
       my $low5=$stat->percentile(5);
       my $up95=$stat->percentile(95);
       my $c = sprintf "%.7f\t%.7f\t%.7f\t%.7f\t%.7f\t[%.7f, %.7f]\t[%.7f, %.7f]\n",$mean,$median,$stddev,$minimum,$maximum,$low5,$up95,$lower,$upper;
       print OUTP "Mean\tMedian\tStdDev\tMin\tMax\t[5%, 95%]\t[25%, 75%]\n$c";
       undef $stat;


sub indels {
    my $arr=$_[0];
    my $leng=$_[1];
    my @groups;

    for (my $i=0;$i<$leng;$i++) {
      if ($arr->[$i]==0) { next; }
      my $sum=0;
      while ($arr->[$i]>0) {
#        if ($arr->[$i]>1) { print STDERR "Error at index $i, number of indels is $arr->[$i] !!! \n"; }
#        $sum++;
        $sum+=$arr->[$i];
        $i++;              }
      push (@groups, $sum);     }
    return (@groups);
}
